#!/usr/bin/perl -w
# Lyrics Manager script for amaroK-1.x
# version 0.2.13 revision 12.09.2009
#
# Copyright 2006-2009 Nikita Melnichenko [http://nikita.melnichenko.name]
# License: GPL-2

# variables that can't be configured through GUI
# relative config path
$config = "lyrics-manager.conf";
# switch for debug (additional print in log)
$DEBUG = 0;

# ========== default settings and constants ==========

%menu_items = (
	"edit" => 'Edit lyrics',
	"export" => 'Export lyrics',
	"import" => 'Import lyrics',
	"erase" => 'Erase lyrics',
	"export_all_playlist" => q|Export all playlist's lyrics|,
	"import_all_playlist" => q|Import all playlist's lyrics|,
	"fill_up_empty_all_playlist" => q|Fill up all empty playlist's lyrics|,
	"erase_all_playlist" => q|Erase all playlist's lyrics|
	);
# ordered keys must be synced with 'keys %menu_items' and used to avoid Tie::IxHash dependency
@menu_items_ordered_keys = ("edit", "export", "import", "erase", "export_all_playlist", "import_all_playlist", "fill_up_empty_all_playlist", "erase_all_playlist");

%default_settings = (
	# editor must be shell-escaped
	"editor" => "kwrite",
	# user's home as a default directory
	"texts_dir" => $ENV{HOME},
	# '<artist> - <title>.txt' file template
	"file_template" => '%a - %t.txt',
	"export_after_edit" => 1,
	"export_empty" => 0,
	"import_empty" => 0,
	"place_texts_with_music" => 0
	);
# all menu items are available by default
foreach my $item (keys %menu_items)
{
	$default_settings{"menu_item_$item"} = 1;
}
# list of all options those are need to be default if not found in config
@not_found_to_default_list = ("file_template", "export_after_edit", "export_empty", "import_empty", "place_texts_with_music");
# also add all menu items to this list
foreach my $item (keys %menu_items)
{
	push (@not_found_to_default_list, "menu_item_$item");
}

# "unknown" indicators for tags
%unknown = (
	"artist" => "Unknown Artist",
	"title" => "Unknown Title",
	"album" => "Unknown Album",
	"track" => "Unknown Track"
	);
# slash replacement for tags as parts of filename
$slash_replacement_for_tags=" ";

# ========== modules ==========

use Fcntl; # for sysopen
use File::Basename; # for dirname
use File::Path; # for mkpath (creating directory trees)

# ========== functions ==========

# ---------- menu ----------

# PARAMS:
# GLOBAL: <$editor, <$texts_dir
sub add_menu_items
{
	foreach $item (@menu_items_ordered_keys)
	{
		my $sh_escaped_item_text = escape_shell_quote ($menu_items{$item});
		system ("dcop amarok script addCustomMenuItem 'Lyrics Manager' '$sh_escaped_item_text'") if $settings{"menu_item_$item"};
	}
	use sigtrap qw(handler sig_handler normal-signals stack-trace error-signals);
}

sub sig_handler
{
	remove_menu_items ();
	exit;
}

sub remove_menu_items
{
	foreach $item (keys %menu_items)
	{
		my $sh_escaped_item_text = escape_shell_quote ($menu_items{$item});
		system ("dcop amarok script removeCustomMenuItem 'Lyrics Manager' '$sh_escaped_item_text'");
	}
}

# ---------- messaging ----------

# PARAMS:
# LOCAL: <$message
sub amarok_show_message
{
	system ('dcop', 'amarok', 'playlist', 'popupMessage', $_[0]);
}

# PARAMS:
# LOCAL: <$message
sub amarok_show_status_message
{
	system ('dcop', 'amarok', 'playlist', 'shortStatusMessage', $_[0]);
}

# PARAMS:
# LOCAL: <$message, >$result
sub strip_html_tags
{
	(my $result = $_[0]) =~ s|</?[a-z]+[^>]*>||g;
	return $result;
}

# PARAMS:
# LOCAL: <$message, <$verbose_level(0 -> all, 1 -> stderr only, 2 -> none), >$status(=true)
sub show_message
{
	my $verbose_level = $_[1] ? $_[1] : 0;
	print STDERR "Lyrics Manager: ".strip_html_tags ($_[0])."\n" if $verbose_level < 2;
	amarok_show_message ('<b>Lyrics Manager:</b> '.$_[0]) if $verbose_level < 1;
	return 1;
}

# ---------- configuration ----------

# adding configuration option guide:
# 1. %default_settings
# 2. read_config (), while section
# 3. configure ()
# 4. write_config ()
# opt. @not_found_to_default_list

# adding menu item guide:
# 1. %menu_items
# 2. @menu_items_ordered_keys
# 3. apply_settings_conditions ()
# 4. item handler in listener section


# set undefined to default and cancel some conflicting features
# PARAMS:
# GLOBAL: <@not_found_to_default_list
# LOCAL: <>%settings
sub apply_settings_conditions
{
	my $settings = $_[0];

	# applying defaults for some settings, if they are not found in config
	foreach my $option (@not_found_to_default_list)
	{
		$settings->{$option} = $default_settings{$option} unless defined ($settings->{$option});
	}

	if (!defined ($settings->{editor}))
	{
		$settings->{menu_item_edit} = 0;
	}
	if (!defined ($settings->{texts_dir}) && !$settings->{place_texts_with_music})
	{
		$settings->{menu_item_export} = 0;
		$settings->{menu_item_import} = 0;
		$settings->{menu_item_export_all_playlist} = 0;
		$settings->{menu_item_import_all_playlist} = 0;
	}
}

# o'reilly recipe 7.3 for substitution tilde in paths
# PARAMS:
# LOCAL: <$path, >$result
sub substitute_tilde
{
	($path = $_[0]) =~ s{^~([^/]*)}{$1 ? (getpwnam($1))[7] : ($ENV{HOME} || $ENV{LOGDIR} || (getpwnam($>))[7])}ex;
	return $path;
}

# PARAMS:
# GLOBAL: <$config, <%default_settings, <%autoconf, <%menu_items, >%settings
# LOCAL: <$configure_status, >$status(true, if ok)
sub read_config
{
	my $messages = "";
	my $configure_status = $_[0];

	if ($configure_status || !open (CONFIG, "< $config"))
	{
		$messages .= "Can't open config file <i>$autoconf{pwd}/$config</i> for read: $!<br>\n" unless $configure_status;
		if (%configured_settings)
		{
			%settings = %configured_settings;
			$messages .= "Configured settings will be used.";
		}
		elsif (%settings)
		{
			$messages .= "Settings will remain the same."
		}
		else
		{
			%settings = %default_settings;
			$messages .= "Default settings will be used.";
		}
		show_message ($messages) if $messages;

		apply_settings_conditions (\%settings);
		return 0;
	}

	%settings = ();
	my $line = 0;
	while (<CONFIG>)
	{
		chomp;
		$line++;

		if ($_ =~ m/^([a-z_]+)=(.*)$/)
		{
			my $option = $1;
			my $value = $2;
			if ($option eq "editor" || $option eq "file_template")
			{
				$settings{$option} = $value;
			}
			elsif ($option eq "texts_dir")
			{
				$value = substitute_tilde ($value);
				if (-d $value)
				{
					$settings{$option} = $value;
				}
				else
				{
					show_message ("Texts directory <i>$value</i> is not a directory!");
				}
			}
			elsif ($option eq "export_after_edit"
				|| $option eq "export_empty"
				|| $option eq "import_empty"
				|| $option eq "place_texts_with_music"
				|| grep { "menu_item_$_" eq $option } keys %menu_items)
			{
				if ($value ne "0" && $value ne "1")
				{
					$messages .= "line $line: error: bad value for option <i>$option</i> (set to 0)<br>\n";
					$value = 0;
				}
				$settings{$option} = $value;
			}
			else
			{
				$messages .= "line $line: warning: unknown option <i>$option</i> (ignored)<br>\n";
			}
		}
		elsif ($_ =~ m/^#/)
		{
			# just a comment, nothing to do
		}
		else
		{
			$messages .= "line $line: warning: unrecognized sequence (ignored)<br>\n";
		}
	}
	show_message ("config parser messages:<br>\n".$messages) if $messages;

	close (CONFIG) or show_message ("Can't close config file <i>$autoconf{pwd}/$config</i> after reading: $!");

	apply_settings_conditions (\%settings);
	return 1;
}

# gather infomation about system components
# PARAMS:
# GLOBAL: >%autoconf
sub autoconf ()
{
	%autoconf = ();

	`which kdialog`; # system call doesn't work here
	$autoconf{kdialog} = !$?;

	$autoconf{pwd} = `pwd`;
	chomp $autoconf{pwd};
}

# ask questions and write down config based on answers
# PARAMS:
# GLOBAL: <%autoconf, <%settings, >%configured_settings
# LOCAL: >$status(=0 on success, =1 if canceled, =2 if write failed)
sub configure
{
	my $gui_configuration = $autoconf{kdialog};

	if ($gui_configuration)
	{
		%configured_settings = ();
		%base_settings = %settings ? %settings : %default_settings;

		my $command;
		my @result;

		# selecting menu items and features
		$command = "kdialog --icon amarok --caption 'Lyrics Manager: Select features'";
		$command .= " --separate-output --checklist 'Activate menu items and features:'";
		foreach my $item (@menu_items_ordered_keys)
		{
			my $sh_escaped_item_text = escape_shell_quote ($menu_items{$item});
			$command .= " 'menu_item_$item' 'Menu Item: $sh_escaped_item_text' ".($base_settings{"menu_item_$item"} ? "on" : "off");
		}
		$command .= " 'export_after_edit' 'Automatic export after edit' ".($base_settings{export_after_edit} ? "on" : "off");
		$command .= " 'export_empty' 'Force empty lyrics export' ".($base_settings{export_empty} ? "on" : "off");
		$command .= " 'import_empty' 'Force empty lyrics import' ".($base_settings{import_empty} ? "on" : "off");
		$command .= " 'place_texts_with_music' 'Place text with music files (instead of texts dir)' ".($base_settings{place_texts_with_music} ? "on" : "off");

		@result = command_output_to_array ($command);
		return 1 if $?;

		foreach my $item (keys %menu_items)
		{
			$configured_settings{"menu_item_$item"} = (grep { $_ eq "menu_item_$item" } @result) ? 1 : 0;
		}
		$configured_settings{export_after_edit} = (grep { $_ eq "export_after_edit" } @result) ? 1 : 0;
		$configured_settings{export_empty} = (grep { $_ eq "export_empty" } @result) ? 1 : 0;
		$configured_settings{import_empty} = (grep { $_ eq "import_empty" } @result) ? 1 : 0;
		$configured_settings{place_texts_with_music} = (grep { $_ eq "place_texts_with_music" } @result) ? 1 : 0;

		# selecting editor
		if ($configured_settings{"menu_item_edit"})
		{
			my $editor = $default_settings{editor};
			$editor = $base_settings{editor} if defined ($base_settings{editor});
			my $sh_escaped_editor = escape_shell_quote ($editor);
			$editor = `kdialog --icon amarok --caption 'Lyrics Manager: Enter editor' --inputbox 'Enter your preffered editor:' '$sh_escaped_editor'`;
			chomp $editor;
			$configured_settings{editor} = $editor if $editor;
		}

		my $need_import_export = $configured_settings{"menu_item_export"}
			|| $configured_settings{"menu_item_import"}
			|| $configured_settings{"menu_item_export_all_playlist"}
			|| $configured_settings{"menu_item_import_all_playlist"};

		# selecting texts directory
		if ($need_import_export && !$configured_settings{"place_texts_with_music"})
		{
			my $texts_dir = $default_settings{texts_dir};
			$texts_dir = $base_settings{texts_dir} if defined ($base_settings{texts_dir});
			my $sh_escaped_texts_dir = escape_shell_quote ($texts_dir);
			$texts_dir = `kdialog --icon amarok --caption 'Lyrics Manager: Select texts directory' --getexistingdirectory '$sh_escaped_texts_dir'`;
			chomp $texts_dir;
			$configured_settings{texts_dir} = $texts_dir if $texts_dir;
		}
		else
		{
			# store old user's texts dir
			$configured_settings{texts_dir} = $settings{texts_dir} if defined ($settings{texts_dir});
		}

		# selecting template
		if ($need_import_export)
		{
			my $file_template = $default_settings{file_template};
			$file_template = $base_settings{file_template} if defined ($base_settings{file_template});
			my $sh_escaped_file_template = escape_shell_quote ($file_template);
			$file_template = `kdialog --icon amarok --caption 'Lyrics Manager: Edit file template' --inputbox 'Placeholders:
	%t	->	title
	%a	->	artist
	%l	->	album
	%r	->	track
	%f	->	file name w/o ext
	%F	->	file name
	%%	->	%' '$sh_escaped_file_template'`;
			chomp $file_template;
			$configured_settings{file_template} = $file_template if $file_template;
		}

		# restoring conditions
		apply_settings_conditions (\%configured_settings);
	}
	else
	{
		show_message ("KDialog not found, thus GUI configuration is not accessible. You can manually edit simple configuration file <i>$autoconf{pwd}/$config</i>");
	}

	# config writing (if used GUI configuration or if no config at all - to create a default config)
	if ($gui_configuration || !(-f $config))
	{
		return write_config () ? 0 : 2;
	}
	return 0;
}

# PARAMS:
# GLOBAL: <$config, <%autoconf, <%default_settings, <%configured_settings, <%menu_items, <@menu_items_ordered_keys
# LOCAL: >$status(true, if ok)
sub write_config
{
	open (CONFIG, "> $config") or show_message ("Can't open config file <i>$autoconf{pwd}/$config</i> for write: $!") && return 0;
	print CONFIG "# Lyrics Manager configuration\n";
	print CONFIG "# preffered editor [string]\n";
	print CONFIG ($configured_settings{editor} ? "editor=$configured_settings{editor}\n" : "#editor=$default_settings{editor}\n");
	print CONFIG "# texts directory [string]\n";
	print CONFIG ($configured_settings{texts_dir} ? "texts_dir=$configured_settings{texts_dir}\n" : "#texts_dir=$default_settings{texts_dir}\n");
	print CONFIG "# file template [string]\n";
	print CONFIG ($configured_settings{file_template} ? "file_template=$configured_settings{file_template}\n"
		: "#file_template=$default_settings{file_template}\n");
	print CONFIG "# automatic export after edit [flag: 0 or 1]\n";
	print CONFIG "export_after_edit=".($configured_settings{export_after_edit} ? 1 : 0)."\n";
	print CONFIG "# force empty lyrics export [flag: 0 or 1]\n";
	print CONFIG "export_empty=".($configured_settings{export_empty} ? 1 : 0)."\n";
	print CONFIG "# force empty lyrics import [flag: 0 or 1]\n";
	print CONFIG "import_empty=".($configured_settings{import_empty} ? 1 : 0)."\n";
	print CONFIG "# place text with music files (instead of texts dir) [flag: 0 or 1]\n";
	print CONFIG "place_texts_with_music=".($configured_settings{place_texts_with_music} ? 1 : 0)."\n";
	print CONFIG "# menu configuration\n";
	foreach my $item (@menu_items_ordered_keys)
	{
		print CONFIG "# menu item $menu_items{$item} [flag: 0 or 1]\n";
		print CONFIG "menu_item_$item=".($configured_settings{"menu_item_$item"} ? 1 : 0)."\n";
	}
	close (CONFIG) or show_message ("Can't close config file <i>$autoconf{pwd}/$config</i> after writing: $!") && return 0;
	return 1;
}

# ---------- file processing ----------

# save output of command to array with chomping; command must be shell-escaped
# PARAMS:
# LOCAL: <$command, >@array
sub command_output_to_array
{
	my @result;

	open (HANDLE, "$_[0] |") or return undef;
	while (<HANDLE>)
	{
		chomp;
		push (@result, $_);
	}
	close (HANDLE) or return undef;

	return @result;
}

# read text file to array with chomping
# PARAMS:
# LOCAL: <$url, >@array
sub read_text_file_to_array
{
	my @result;

	sysopen (HANDLE, $_[0], O_RDONLY) or return undef;
	while (<HANDLE>)
	{
		chomp;
		push (@result, $_);
	}
	close (HANDLE) or return undef;

	return @result;
}

# write array to text file
# PARAMS:
# LOCAL: <$url, <@array, >$status(true, if ok)
sub write_text_file_from_array
{
	my @array = @{$_[1]};
	my $path = $_[0];

	# try to create a path
	my $dir = dirname ($path);
	eval { mkpath ($dir) };
	if ($@)
	{
		return 0;
	}

	sysopen (FILE, $path, O_WRONLY | O_TRUNC | O_CREAT) or return 0;
	foreach my $line (@array)
	{
		print FILE "$line\n";
	}
	close (FILE) or return 0;

	return 1;
}

# test result of array-return function for undef
# PARAMS:
# LOCAL: <@array, >$result_bool
sub bad_array_result
{
	return ($#{$_[0]} == 0 && !defined (${$_[0]}[0]));
}

# ---------- localized amaroK interface ----------

# escape 's for shell
# PARAMS:
# LOCAL: <$string, >$result
sub escape_shell_quote
{
	(my $result = $_[0]) =~ s/'/'\\''/g;
	return $result;
}

# escape 's for sql query
# PARAMS:
# LOCAL: <$string, >$result
sub escape_sql_quote
{
	(my $result = $_[0]) =~ s/'/''/g;
	return $result;
}

# read select query result to array of strings
# PARAMS:
# LOCAL: <$query, >@array
sub amarok_db_select_query
{
	show_message ("DEBUG [SQL query]: $_[0]", 1) if ($DEBUG);
	my $sh_escaped_query = escape_shell_quote ($_[0]);
	my @result = command_output_to_array ("dcop amarok collection query '$sh_escaped_query'");
	show_message ("DEBUG [DB result]:\n----- begin output -----\n".join ("\n", @result)."\n----- end output -----\n", 1) if ($DEBUG);
	return @result;
}

# read tag from collection database
# ! assumption: after successful amarok_collection_search
# PARAMS:
# GLOBAL: <%file_info
# LOCAL: <$tag_name("title", "artist", "album", "track"), >$result
sub amarok_db_get_tag
{
	my $tag_name = $_[0];
	my $sql_escaped_url = escape_sql_quote ($file_info{relative_url});
	my $query;
	my @db_result;
	my $result = "";

	if ($sql_escaped_url eq "")
	{
		return "";
	}

	if ($tag_name eq "title")
	{
		$query = "select tags.title from tags where tags.url = '$sql_escaped_url'"
			." and tags.deviceid = $file_info{device_id}";
		$result = $db_result[0] if (@db_result = amarok_db_select_query ($query));
	}

	if ($tag_name eq "artist")
	{
		$query = "select artist.name from tags, artist where tags.url = '$sql_escaped_url'"
			." and tags.deviceid = $file_info{device_id} and tags.artist = artist.id";
		$result = $db_result[0] if (@db_result = amarok_db_select_query ($query));
	}

	if ($tag_name eq "album")
	{
		$query = "select album.name from tags, album where tags.url = '$sql_escaped_url'"
			." and tags.deviceid = $file_info{device_id} and tags.album = album.id";
		$result = $db_result[0] if (@db_result = amarok_db_select_query ($query));
	}

	if ($tag_name eq "track")
	{
		$query = "select tags.track from tags where tags.url = '$sql_escaped_url'"
			." and tags.deviceid = $file_info{device_id}";
		$result = $db_result[0] if (@db_result = amarok_db_select_query ($query));
	}

	return $result;
}

# PARAMS:
# LOCAL: <$url, >@array
sub amarok_get_lyrics_by_url
{
	my $sh_escaped_url = escape_shell_quote ($_[0]);
	return command_output_to_array ("dcop amarok player lyricsByPath '$sh_escaped_url'");
}

# PARAMS:
# LOCAL: <$url, <@array, >$status(true, if ok)
sub amarok_set_lyrics_by_url
{
	my $sh_escaped_lyrics = escape_shell_quote (join ("\n", @{$_[1]}));
	# can't cache $current_url to reduce dcop calls during mass operations, because it could be changed every moment
	my $current_url = `dcop amarok player path`; chomp $current_url;
	if ($current_url eq $_[0])
	{
		# this call should set lyrics and refresh lyrics tab
		# thanks for help to: panzi at KDE-Apps.org
		return !system ("dcop amarok contextbrowser showLyrics '$sh_escaped_lyrics'");
	}
	else
	{
		my $sh_escaped_url = escape_shell_quote ($_[0]);
		return !system ("dcop amarok player setLyricsByPath '$sh_escaped_url' '$sh_escaped_lyrics'");
	}
}

# PARAMS:
# LOCAL: <$encoded_url, >$decoded_url
sub decode_url
{
	(my $url = $_[0]) =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
	return $url;
}

# retrieve file urls of current playlist
# ! hardly used that open-tag 'item' written in one line
# PARAMS:
# LOCAL: <$verbose_level, >@urls(empty, if failed)
sub amarok_playlist_read_urls
{
	my $xml_location = `dcop amarok playlist saveCurrentPlaylist`; chomp $xml_location;
	my @xml_contents = read_text_file_to_array ($xml_location);
	if (bad_array_result (\@xml_contents))
	{
		show_message ("Can't read playlist file <i>$xml_location</i>: $!", $_[0]);
		return ();
	}

	my @urls;
	foreach $xml_line (@xml_contents)
	{
		if ($xml_line =~ m|<item [^>]*url="file://([^"]*)"|)
		{
			push (@urls, decode_url ($1));
		}
	}

	return @urls;
}

# check guess of relative url and device id 
# PARAMS:
# GLOBAL: >%file_info
# LOCAL: <$relative_url, <$device_id, >$result_bool
sub amarok_collection_search_check_guess
{
	my $relative_url = $_[0];
	my $device_id = $_[1];

	show_message ("DEBUG [collection search]: new device id guess = $device_id", 1) if ($DEBUG);

	my $sql_escaped_url = escape_sql_quote ($relative_url);
	my $query = "select tags.url from tags where tags.url = '$sql_escaped_url' and tags.deviceid = $device_id";
	my @db_result = amarok_db_select_query ($query);
	if ($db_result[0])
	{
		show_message ("DEBUG [collection search]: file was found!", 1) if ($DEBUG);
		%file_info = (
			"relative_url" => $relative_url,
			"device_id" => $device_id
			);
		return 1;
	}

	show_message ("DEBUG [collection search]: no entry found", 1) if ($DEBUG);
	return 0;
}

# do collection search for given url trying to split it to device id and relative url
# ! note that this is ad hoc algorithm, because no specs available for database structure
# ! if you know how to do this in the proper way, please, contact me (see README file for contacts)
# thanks for help to: bleearg at KDE-Apps.org
# PARAMS:
# GLOBAL: >%file_info
# LOCAL: <$url, >$result_bool
sub amarok_collection_search
{
	my $url = $_[0];
	show_message ("DEBUG [collection search]: URL = $url", 1) if ($DEBUG);
	%file_info = ();

	# retrieve divices amd mount points
	my @device_list = amarok_db_select_query ('select lastmountpoint, id from devices');

	# try to guess mount point as largest prefix for given url
	my $best_mpoint_guess = "";
	for ($i = 0; $i < $#device_list / 2; $i++)
	{
		if (length ($best_mpoint_guess) < length ($device_list[2*$i]) && ($url =~ m/^$device_list[2*$i]/))
		{
			$best_mpoint_guess = $device_list[2*$i];
			show_message ("DEBUG [collection search]: new mount point guess = $best_mpoint_guess", 1) if ($DEBUG);
		}
	}
	if ($best_mpoint_guess eq "")
	{
		show_message ("DEBUG [collection search]: no mount point guess at all - this could mean that devices table is corrupted!", 1) if ($DEBUG);
		show_message ("DEBUG [collection search]: will try the unknown device id (-1) later!", 1) if ($DEBUG);
	}

	# extract relative url
	$url =~ m/^$best_mpoint_guess(.*)$/;
	my $relative_url = $1;
	# correct relative url to the form, in which it could be written in database
	if ($relative_url =~ m|^/|)
	{
		$relative_url = ".$relative_url";
	}
	else
	{
		$relative_url = "./$relative_url";
	}
	show_message ("DEBUG [collection search]: relative URL = $relative_url", 1) if ($DEBUG);

	# try to find proper device id (as practice say, there could be more than one device for given moint point)
	for ($i = 0; $i < $#device_list / 2; $i++)
	{
		if ($best_mpoint_guess eq $device_list[2*$i])
		{
			if (amarok_collection_search_check_guess ($relative_url, $device_list[2*$i+1]))
			{
				return 1;
			}
		}
	}
	# try the unknown device id (-1)
	if ($best_mpoint_guess eq "")
	{
		if (amarok_collection_search_check_guess ($relative_url, -1))
		{
			return 1;
		}
	}
	# can't find any - may be it's not in collection?
	show_message ("DEBUG [collection search]: file wasn't found!", 1) if ($DEBUG);
	return 0;
}

# ---------- lyrics text processing ----------

# PARAMS:
# LOCAL: <@array, >@result
sub lyrics_cleanup
{
	my @in = @{$_[0]};
	my @out;

	foreach my $line (@in)
	{
		# replace trivial tags
		$line =~ s|&apos;|'|g;
		$line =~ s|&quot;|"|g;
		$line =~ s|&amp;|&|g;
		# clean trailing br-tag with invisible simbols
		$line =~ s|<[ ]*br[ ]*/?[ ]*>[ \t\r]*$||g;
		$line =~ s|[ \t\r]*$||g;
		# split line to substrings depending on internal br-tags
		$line =~ s|[ \t\r]*<[ ]*br[ ]*/?[ ]*>|\n|g;
		my @line_split = split (/\n/, $line);
		@line_split = ("") unless $line; # split returns empty array for empty string
		push (@out, @line_split);
	}

	# clean empty leading and trailing lines
	# search
	my $head; my $tail;
	for ($head = 0; $head <= $#out && $out[$head] eq ""; $head++) {}
	return () if $head > $#out; # empty lyrics
	for ($tail = $#out; $tail >= 0 && $out[$tail] eq ""; $tail--) {}
	# rearrangement
	my $length = $tail - $head + 1;
	for (my $i = 0; $i < $length; $i++)
	{
		$out[$i] = $out[$head + $i];
	}
	$#out = $length - 1;

	return @out;
}

# modify lyrics from amarok state to text state
# PARAMS:
# LOCAL: <@array, >@result
sub lyrics_amarok_to_text
{
	my @in = @{$_[0]};
	my @out;

	foreach my $line (@in)
	{
		$line =~ s|</?lyrics[^>]*>||g;
		$line =~ s|<\?xml[^?]*\?>||g;
		push (@out, $line);
	}

	return lyrics_cleanup (\@out);
}

# modify lyrics from text state to amarok state
# PARAMS:
# GLOBAL: <%autoconf, <%unknown
# LOCAL: <@array, <$artist, <$title, <$no_cleanup, >@result
sub lyrics_text_to_amarok
{
	my @in = $_[3] ? @{$_[0]} : lyrics_cleanup ($_[0]);
	my @out;

	my $artist = $_[1];
	my $title = $_[2];

	# make proper xml strings
	$artist =~ s/'/&apos;/g;
	$title  =~ s/'/&apos;/g;
	$artist =~ s/"/&quot;/g;
	$title  =~ s/"/&quot;/g;
	$artist =~ s/&/&amp;/g;
	$title  =~ s/&/&amp;/g;

	@out = @in;
	if ($#out >= 0)
	{
		$out[0] = "<lyrics"
			.($artist && $artist ne $unknown{artist} ? " artist=\"$artist\"" : "")
			.($title && $title ne $unknown{title} ? " title=\"$title\"" : "")
			.">"
			.$out[0];
		$out[$#out] = $out[$#out]."</lyrics>";
	}

	return @out;
}

# ---------- actions ----------

# PARAMS:
# LOCAL: <@array, <$verbose_level, >@result
sub lyrics_edit
{
	my @lyrics = @{$_[0]};
	my $verbose_level = $_[1];

	my $temp_file = `mktemp -t lyrics-temp-XXXXXX`;
	if ($? != 0)
	{
		show_message ("Can't create temporary filename: $!", $verbose_level);
		return undef;
	}
	chomp $temp_file;

	write_text_file_from_array ($temp_file, \@lyrics) or show_message ("Can't write to temporary file: $!", $verbose_level) && return undef;

	my $sh_escaped_temp_file = escape_shell_quote ($temp_file);
	# neither system call nor $? don't work here
	$status = `$settings{editor} '$sh_escaped_temp_file' 1>/dev/null 2>/dev/null; echo \$?`;
	if ($status == 127)
	{
		show_message ("Your editor <i>$settings{editor}</i> not found!", $verbose_level);
		unlink ($temp_file);
		return undef;
	}

	@lyrics = read_text_file_to_array ($temp_file);
	if (bad_array_result (\@lyrics))
	{
		show_message ("Can't read temporary file: $!", $verbose_level);
		unlink ($temp_file);
		return undef;
	}
	unlink ($temp_file);

	return @lyrics;
}

# calculate export and import information
# PARAMS:
# GLOBAL: <%settings, <%unknown, <%file_info
# LOCAL: <$url, >%result
sub external_params
{
	my $url = $_[0];
	my $tag;
	my %result;

	# required tags
	$result{artist} = ($tag = amarok_db_get_tag ("artist")) ? $tag : $unknown{artist};
	$result{title} = ($tag = amarok_db_get_tag ("title")) ? $tag : $unknown{title};

	my $template = $settings{file_template};
	my $file = "";
	while ($template =~ m/^([^%]*)%(.)(.*)$/)
	{
		my $tag = "";
		if ($2 eq "t")
		{
			$tag = $result{title};
		}
		elsif ($2 eq "a")
		{
			$tag = $result{artist};
		}
		elsif ($2 eq "l")
		{
			$result{album} = ($tag = amarok_db_get_tag ("album")) ? $tag : $unknown{album} if !$result{album};
			$tag = $result{album};
		}
		elsif ($2 eq "r")
		{
			$result{track} = ($tag = amarok_db_get_tag ("track")) ? $tag : $unknown{track} if !$result{track};
			# Because of the standard sorting issues, we prefix 0 to a number, if it is 9 or smaller
			$tag = ($result{track} <= 9) ? "0$result{track}" : $result{track};
		}
		elsif ($2 eq "f")
		{
			$result{orig_file_prefix} = basename ($url);
			$result{orig_file_prefix} =~ s/\.[^.]*$//;
			$tag = $result{orig_file_prefix};
		}
		elsif ($2 eq "F")
		{
			$result{orig_file_name} = basename ($url);
			$tag = $result{orig_file_name};
		}
		elsif ($2 eq "%")
		{
			$tag = "%";
		}
		# append unaltered content before tag to file name
		$file .= $1;
		# set template to unhandled part of it
		$template = $3;
		# replace bad chars in tag
		$tag =~ s|/|$slash_replacement_for_tags|g;
		# append tag to file name
		$file .= $tag;
	}
	# append tail to file name
	$file .= $template;
	$result{filename} = $file;

	if ($settings{place_texts_with_music})
	{
		$result{path} = dirname ($url)."/$result{filename}";
	}
	else
	{
		$result{path} = "$settings{texts_dir}/$result{filename}";
	}
	return %result;
}

# lyrics must be cleaned up before
# PARAMS:
# GLOBAL: <%settings
# LOCAL: <%params, <@lyrics, <$verbose_level, >$status(=0 on success, =1 on empty lyrics, =2 on error)
sub lyrics_export
{
	my %params = %{$_[0]};
	my $lyrics = $_[1];
	my $verbose_level = $_[2];

	if (!$settings{export_empty} && !@$lyrics)
	{
		show_message ("Lyrics are empty. You must enable empty lyrics exporting to proceed.", $verbose_level);
		return 1;
	}

	if (write_text_file_from_array ($params{path}, $lyrics))
	{
		show_message ("File <i>$params{filename}</i> saved.", $verbose_level);
	}
	else
	{
		show_message ("Can't write to file <i>$params{filename}</i>: $!", $verbose_level);
		return 2;
	}

	return @$lyrics ? 0 : 1;
}

# PARAMS:
# GLOBAL: <%settings
# LOCAL: <$url, <$verbose_level
sub action_edit
{
	my $url = $_[0];
	my $verbose_level = $_[1];
	if (!amarok_collection_search ($url))
	{
		show_message ("Music file '$url' is not in your Collection!", $verbose_level);
		return;
	}

	my @lyrics = amarok_get_lyrics_by_url ($url);
	if (bad_array_result (\@lyrics))
	{
		show_message ("Can't get lyrics from amaroK!", $verbose_level);
		return;
	}
	@lyrics = lyrics_amarok_to_text (\@lyrics);

	@lyrics = lyrics_edit (\@lyrics);
	return if bad_array_result (\@lyrics);

	@lyrics = lyrics_cleanup (\@lyrics);

	my %params = external_params ($url);
	if ($settings{export_after_edit})
	{
		lyrics_export (\%params, \@lyrics, $verbose_level);
	}

	@lyrics = lyrics_text_to_amarok (\@lyrics, $params{artist}, $params{title});
	amarok_set_lyrics_by_url ($url, \@lyrics) or show_message ("Can't set lyrics to amaroK!", $verbose_level);
}

# PARAMS:
# LOCAL: <$url, <$verbose_level, >$status(=0 on success, =1 on empty lyrics, =2 on error, =3 on non-collection url)
sub action_export
{
	my $url = $_[0];
	my $verbose_level = $_[1];
	if (!amarok_collection_search ($url))
	{
		show_message ("Music file '$url' is not in your Collection!", $verbose_level);
		return 3;
	}

	my %params = external_params ($url);
	my @lyrics = amarok_get_lyrics_by_url ($url);
	if (bad_array_result (\@lyrics))
	{
		show_message ("Can't get lyrics from amaroK!", $verbose_level);
		return 2;
	}

	@lyrics = lyrics_amarok_to_text (\@lyrics);
	return lyrics_export (\%params, \@lyrics, $verbose_level);
}

# PARAMS:
# GLOBAL: <%settings
# LOCAL: <$url, <$verbose_level, >$status(=0 on success, =1 on empty lyrics, =2 on error, =3 on non-collection url, =4 not found)
sub action_import
{
	my $url = $_[0];
	my $verbose_level = $_[1];
	if (!amarok_collection_search ($url))
	{
		show_message ("Music file '$url' is not in your Collection!", $verbose_level);
		return 3;
	}

	my %params = external_params ($url);
	-f $params{path} or show_message ("File <i>$params{filename}</i> not found in texts directory.", $verbose_level) && return 4;
	my @lyrics = read_text_file_to_array ($params{path});
	if (bad_array_result (\@lyrics))
	{
		show_message ("Can't read file <i>$params{filename}</i>: $!", $verbose_level);
		return 2;
	}

	@lyrics = lyrics_text_to_amarok (\@lyrics, $params{artist}, $params{title});
	if (!$settings{import_empty} && !@lyrics)
	{
		show_message ("Lyrics are empty. You must enable empty lyrics importing to proceed.", $verbose_level);
		return 1;
	}

	if (amarok_set_lyrics_by_url ($url, \@lyrics))
	{
		show_message ("File <i>$params{filename}</i> loaded.", $verbose_level);
	}
	else
	{
		show_message ("Can't set lyrics to amaroK!", $verbose_level);
		return 2;
	}

	return @lyrics ? 0 : 1;
}

# PARAMS:
# GLOBAL: <%settings
# LOCAL: <@urls
sub action_mass_export
{
	my @urls = @{$_[0]};
	my ($total, $empty, $errors, $non_db, $ignored) = (0, 0, 0, 0, 0);

	foreach my $url (@urls)
	{
		amarok_show_status_message ("Processing $url...");
		my $status = action_export ($url, 1);
		$total++;
		$empty++ if $status == 1;
		$errors++ if $status == 2;
		$non_db++ if $status == 3;
		$ignored++ if $status == 3 || $status == 1 && !$settings{export_empty};
	}
	my $normal = $total - $empty - $errors - $non_db;
	my $saved = $total - $ignored - $errors;
	my $stats_message = "Total: $total\n<ul><li>Normal: $normal</li>\n<li>Empty: $empty</li>\n<li>Errors: $errors</li>\n<li>Not in collection: $non_db</li></ul>\nTotal saved: $saved<br>\nTotal ignored: $ignored";

	amarok_show_status_message ("Done. Processed: $total. Ignored: $ignored.");
	show_message ("Export complete.<br>\n<i>Statistics (see log for details):</i><br>\n$stats_message");
}

# PARAMS:
# LOCAL: <@urls
sub action_mass_import
{
	my @urls = @{$_[0]};
	my ($total, $empty, $errors, $non_db, $not_found, $ignored) = (0, 0, 0, 0, 0, 0);

	foreach my $url (@urls)
	{
		amarok_show_status_message ("Processing $url...");
		my $status = action_import ($url, 1);
		$total++;
		$empty++ if $status == 1;
		$errors++ if $status == 2;
		$non_db++ if $status == 3;
		$not_found++ if $status == 4;
		$ignored++ if $status == 3 || $status == 4 || $status == 1 && !$settings{import_empty};
	}
	my $normal = $total - $empty - $errors - $non_db - $not_found;
	my $loaded = $total - $ignored - $errors;
	my $stats_message = "Total: $total\n<ul><li>Normal: $normal</li>\n<li>Empty: $empty</li>\n<li>Not found: $not_found</li>\n<li>Errors: $errors</li>\n<li>Not in collection: $non_db</li></ul>\nTotal loaded: $loaded<br>\nTotal ignored: $ignored";

	amarok_show_status_message ("Done. Processed: $total. Ignored: $ignored.");
	show_message ("Import complete.<br>\n<i>Statistics (see log for details):</i><br>\n$stats_message");
}

# PARAMS:
# GLOBAL: <%settings
# LOCAL: <$url, <$verbose_level, >$status(=0 on non-empty lyrics, =1 on fill, =2 on error, =3 on non-collection url)
sub action_fill_up_empty
{
	my $url = $_[0];
	my $verbose_level = $_[1];
	if (!amarok_collection_search ($url))
	{
		show_message ("Music file '$url' is not in your Collection!", $verbose_level);
		return 3;
	}

	my @lyrics = amarok_get_lyrics_by_url ($url);
	if (bad_array_result (\@lyrics))
	{
		show_message ("Can't get lyrics from amaroK!", $verbose_level);
		return 2;
	}

	@lyrics = lyrics_amarok_to_text (\@lyrics);
	return 0 if (@lyrics);

	my %params = external_params ($url);
	@lyrics = (" ");
	@lyrics = lyrics_text_to_amarok (\@lyrics, $params{artist}, $params{title}, 1);

	if (amarok_set_lyrics_by_url ($url, \@lyrics))
	{
		return 1;
	}
	else
	{
		show_message ("Can't set lyrics to amaroK!", $verbose_level);
		return 2;
	}
}

# fill up empty lyrics with space to avoid automatic fetching
# PARAMS:
# LOCAL: <@urls
sub action_mass_fill_up_empty
{
	my @urls = @{$_[0]};
	my ($total, $filled, $errors, $non_db) = (0, 0, 0, 0);

	foreach my $url (@urls)
	{
		amarok_show_status_message ("Processing $url...");
		my $status = action_fill_up_empty ($url, 1);
		$total++;
		$filled++ if $status == 1;
		$errors++ if $status == 2;
		$non_db++ if $status == 3;
	}
	my $normal = $total - $filled - $errors - $non_db;
	my $stats_message = "Total: $total\n<ul><li>Normal: $normal</li>\n<li>Filled: $filled</li>\n<li>Errors: $errors</li>\n<li>Not in collection: $non_db</li></ul>";

	amarok_show_status_message ("Done. Processed: $total. Filled: $filled.");
	show_message ("Filling complete.<br>\n<i>Statistics (see log for details):</i><br>\n$stats_message");
}

# PARAMS:
# GLOBAL: <%settings
# LOCAL: <$url, <$verbose_level, >$status(=0 on success, =2 on error, =3 on non-collection url)
sub action_erase
{
	my $url = $_[0];
	my $verbose_level = $_[1];
	if (!amarok_collection_search ($url))
	{
		show_message ("Music file '$url' is not in your Collection!", $verbose_level);
		return 3;
	}

	my %params = external_params ($url);
	@lyrics = ();

	if (amarok_set_lyrics_by_url ($url, \@lyrics))
	{
		amarok_show_status_message ("Lyrics erased.") unless $verbose_level;
		return 0;
	}
	else
	{
		show_message ("Can't set lyrics to amaroK!", $verbose_level);
		return 2;
	}
}

# PARAMS:
# LOCAL: <@urls
sub action_mass_erase
{
	my @urls = @{$_[0]};
	my ($total, $errors, $non_db) = (0, 0, 0);

	foreach my $url (@urls)
	{
		amarok_show_status_message ("Processing $url...");
		my $status = action_erase ($url, 1);
		$total++;
		$errors++ if $status == 2;
		$non_db++ if $status == 3;
	}
	my $erased = $total - $errors - $non_db;
	my $stats_message = "Total: $total\n<ul><li>Erased: $erased</li>\n<li>Errors: $errors</li>\n<li>Not in collection: $non_db</li></ul>";

	amarok_show_status_message ("Done. Processed: $total. Erased: $erased.");
	show_message ("Erasing complete.<br>\n<i>Statistics (see log for details):</i><br>\n$stats_message");
}

# ========== initialization ==========

autoconf ();

if (!(-f $config))
{
	show_message ("This is the first start of Lyrics Manager. You need to configure the script.");
	configure ();
}

read_config ();
remove_menu_items (); # cleanup (menu could remain, if script was killed and crash handler was not launched)
add_menu_items ();

# ========== listener ==========

while (<STDIN>)
{
	$message = $_; chomp $message;
	($message =~ m/^([A-Za-z]*)/) && ($message_type = $1) or next;

	if ($message_type eq "configure")
	{
		remove_menu_items ();
		read_config (configure ());
		add_menu_items ();
		next;
	}
	if ($message_type eq "customMenuClicked" && $message =~ m|^customMenuClicked: Lyrics Manager ([A-Za-z' ]*) file://(.*)$|)
	{
		$type = $1;
		$url = decode_url ($2);

		if ($type eq $menu_items{'edit'})
		{
			action_edit ($url);
			next;
		}
		elsif ($type eq $menu_items{'export'})
		{
			action_export ($url);
			next;
		}
		elsif ($type eq $menu_items{'import'})
		{
			action_import ($url);
			next;
		}
		elsif ($type eq $menu_items{'erase'})
		{
			action_erase ($url);
			next;
		}
		elsif ($type eq $menu_items{'export_all_playlist'})
		{
			my @urls = amarok_playlist_read_urls ();
			action_mass_export (\@urls);
			next;
		}
		elsif ($type eq $menu_items{'import_all_playlist'})
		{
			my @urls = amarok_playlist_read_urls ();
			action_mass_import (\@urls);
			next;
		}
		elsif ($type eq $menu_items{'fill_up_empty_all_playlist'})
		{
			my @urls = amarok_playlist_read_urls ();
			action_mass_fill_up_empty (\@urls);
			next;
		}
		elsif ($type eq $menu_items{'erase_all_playlist'})
		{
			my @urls = amarok_playlist_read_urls ();
			action_mass_erase (\@urls);
			next;
		}
	}
}
